"""
Utilities for (de)serializing binary strings.
"""

import logging

logger = logging.getLogger(__name__)

import fractions
import struct
from typing import Optional


def bytes_to_string(data: bytes) -> str:
    """Convert a binary string to a utf-8 string."""
    return data.decode("utf-8")


def string_to_bytes(string: str) -> bytes:
    """Convert a utf-8 string to a binary string."""
    return string.encode("utf-8")


def bytes_to_bitvector(bitvector_bytes: bytes) -> list[bool]:
    """Convert a bytestring representing a bitvector into a list of booleans."""
    return [(byte >> bit) & 1 == 1 for byte in bitvector_bytes for bit in range(8)]


def bitvector_to_bytes(bitvector: list[bool]) -> bytes:
    """Convert a list of booleans representing a bitvector into a bytestring."""
    byte_array = bytearray()
    for i in range(0, len(bitvector), 8):
        byte = sum((1 << j) for j in range(8) if i + j < len(bitvector) and bitvector[i + j])
        byte_array.append(byte)
    return bytes(byte_array)


def endianness_to_struct_format(little_endian: bool) -> str:
    """Convert endianness flag to a formatting string for struct."""
    return "<" if little_endian else ">"

def assert_key_in_dict(table: dict, key, desc):
    if key not in table:
        raise ValueError(f"{desc} must be one of {list(table.keys())} but is {key}")

def value_type_to_struct_format(value_type: str) -> str:
    """Convert a value type to a formatting string for struct."""
    table = {
        "int32": "i",
        "int64": "q",
        "uint32": "I",
        "uint64": "Q",
        "double": "d",
    }
    assert_key_in_dict(table, value_type, "value type")
    return table[value_type]



def bytes_to_value(
    data: bytes, value_type: str, little_endian: bool = True
) -> int | float | fractions.Fraction | str | tuple:
    """
    Convert a binary string to a single value of the given type.
    :param value_type: string or one of {int32|uint32|int64|uint64|double|rational}[-interval]
    :return: a pair of left and right values for interval types, or a single value otherwise
    """

    if value_type == "string":
        return bytes_to_string(data)

    if "-interval" in value_type:
        assert len(data) % 2 == 0, "interval data must have even length"
        mid = len(data) // 2
        base_value_type = value_type.replace("-interval", "")
        left = bytes_to_value(data[:mid], base_value_type)
        right = bytes_to_value(data[mid:], base_value_type)
        return (left, right)

    if value_type == "rational":
        assert len(data) % 2 == 0, "rational data must have even length"
        mid = len(data) // 2
        byteorder = "little" if little_endian else "big"
        numerator = int.from_bytes(data[:mid], byteorder=byteorder, signed=True)
        denominator = int.from_bytes(data[mid:], byteorder=byteorder, signed=False)
        return fractions.Fraction(numerator, denominator)

    assert value_type in ["int32", "uint32", "int64", "uint64", "double"], "unexpected value type"
    type_format = value_type_to_struct_format(value_type)
    endian_format = endianness_to_struct_format(little_endian)
    return struct.unpack(f"{endian_format}{type_format}", data)[0]


def normalize_rational(value: fractions.Fraction) -> fractions.Fraction:
    """Ensure that the denominator of a rational number is non-negative."""
    if value.denominator < 0:
        value = fractions.Fraction(-value.numerator, -value.denominator)
    return value


def standard_value_type_size(value_type: str) -> int:
    """Return the number of bytes needed to represent a value of the given type."""
    if "-interval" in value_type:
        base_value_type = value_type.replace("-interval", "")
        return 2 * standard_value_type_size(base_value_type)
    if value_type == "rational":
        # for rationals, the standard size is 8+8 bytes
        return standard_value_type_size("int64") + standard_value_type_size("uint64")
    table = {
        "int32": 4,
        "int64": 8,
        "uint32": 4,
        "uint64": 8,
        "double": 8,
    }
    assert_key_in_dict(table, value_type, "value type")
    return table[value_type]


def integer_value_size(value: int) -> int:
    """Return the number of bytes needed to represent an integer value, rounded up to the nearest multiple of 8."""
    num_bytes = (value.bit_length() + 7) // 8
    if num_bytes % 8 != 0:
        num_bytes += 8 - (num_bytes % 8)
    return num_bytes


def rational_value_size(value: fractions.Fraction) -> int:
    """Return the number of bytes needed to represent a rational value."""
    return 2*max(integer_value_size(value.numerator), integer_value_size(value.denominator))


def rational_to_bytes(value: fractions.Fraction, max_integer_size: Optional[int] = None, little_endian: bool = True) -> bytes:
    """
    Convert a rational number to a bytestring. Both numberator and denominator are encoded as signed and unsigned integers, respectively, and have the same size.
    :param max_integer_size: (optional) maximum size in bytes for numerator and denominator; if not provided, the size is determined automatically
    """
    value = normalize_rational(value)
    if max_integer_size is None:
        max_integer_size = rational_value_size(value) // 2
    byteorder = "little" if little_endian else "big"
    numerator_bytes = value.numerator.to_bytes(max_integer_size, byteorder=byteorder, signed=True)
    denominator_bytes = value.denominator.to_bytes(max_integer_size, byteorder=byteorder, signed=False)
    return numerator_bytes + denominator_bytes


def value_to_bytes(
    value: int | float | tuple | fractions.Fraction | str, value_type: str, little_endian: bool = True
) -> bytes:
    """
    Convert a value of a given type to a bytestring.
    :param value_type: either string or one of {int32|uint32|int64|uint64|double|rational}[-interval]
    """

    if value_type == "string":
        assert isinstance(value, str), "string value must be a str"
        return string_to_bytes(value)

    if value_type == "rational":
        assert isinstance(value, fractions.Fraction), "rational value must be a Fraction"
        return rational_to_bytes(value, little_endian=little_endian)

    if "-interval" in value_type:
        assert isinstance(value, tuple) and len(value) == 2, "interval value must be a tuple of two values"
        base_value_type = value_type.replace("-interval", "")
        if base_value_type != "rational":
            lower = value_to_bytes(value[0], base_value_type, little_endian)
            upper = value_to_bytes(value[1], base_value_type, little_endian)
        else:
            assert all(isinstance(v, fractions.Fraction) for v in value), "expected a pair of fractions"
            # ensure both rationals use the same size for numerator and denominator
            max_integer_size = max(rational_value_size(v) for v in value) // 2
            lower = rational_to_bytes(value[0], max_integer_size, little_endian)
            upper = rational_to_bytes(value[1], max_integer_size, little_endian)
        return lower + upper

    assert value_type in ["int32", "uint32", "int64", "uint64", "double"], "unexpected value type"
    type_format = value_type_to_struct_format(value_type)
    endian_format = endianness_to_struct_format(little_endian)
    return struct.pack(f"{endian_format}{type_format}", value)


def bytes_into_chunk_ranges(data: bytes, chunk_ranges: list[tuple[int, int]]) -> list[bytes]:
    """Split bytestring into chunks according to chunk ranges."""
    return [data[start:end] for start, end in chunk_ranges]


def bytes_into_num_chunks(data: bytes, num_chunks: int) -> list[bytes]:
    """Split bytestring into evenly sized chunks."""
    assert num_chunks > 0, "num_chunks must be a positive number"
    assert len(data) % num_chunks == 0, "len(data) must be divisible by num_chunks when item_ranges are not provided"
    chunk_size = len(data) // num_chunks
    chunk_ranges = [(i * chunk_size, (i + 1) * chunk_size) for i in range(num_chunks)]
    return bytes_into_chunk_ranges(data, chunk_ranges)


def bytes_to_vector(
    data: bytes, value_type: str, chunk_ranges: Optional[list[tuple[int, int]]] = None, little_endian: bool = True
) -> list:
    """
    Decode a binary string as a list of numbers.
    :param value_type: vector element type, either bool, string or one of {int32|uint32|int64|uint64|double|rational}[-interval]
    :param chunk_ranges: (optional) chunk ranges to split the data into
    :param little_endian: if True, the binary string is interpreted as little-endian
    """
    if value_type == "bool":
        assert little_endian, "big-endianness for bitvectors is not implemented"
        return bytes_to_bitvector(data)

    if len(data) == 0:
        return []

    if chunk_ranges is None:
        chunk_size = standard_value_type_size(value_type)
        assert len(data) % chunk_size == 0, "len(data) must be divisible by the size of the value type"
        num_chunks = len(data) // chunk_size
        chunks = bytes_into_num_chunks(data, num_chunks)
        return [bytes_to_value(chunk, value_type, little_endian) for chunk in chunks]

    return [bytes_to_value(data[start:end], value_type, little_endian) for start, end in chunk_ranges]


def vector_to_bytes(
    vector: list, value_type: str, little_endian: bool = True
) -> tuple[bytes, Optional[list[tuple[int, int]]]]:
    """Encode a list of values as a binary string.
    :param value_type: vector element type, either bool, string or {int32|uint32|int64|uint64|double|rational}[-interval]
    :return: encoded binary string
    :return: (optional) chunk ranges if non-trivial splitting is needed to split the resulting bytestring into chunks, e.g. for strings or non-standard rationals
    """
    if len(vector) == 0:
        logger.warning("converting empty vector to bytes")
        return (b"", None)

    if value_type == "bool":
        assert little_endian, "big-endianness for bitvectors is not implemented"
        return (bitvector_to_bytes(vector), None)

    chunks = [value_to_bytes(item, value_type, little_endian) for item in vector]
    chunk_ranges = None
    if value_type == "string" or any(len(chunk) != standard_value_type_size(value_type) for chunk in chunks):
        chunk_ranges = []
        current_pos = 0
        for chunk in chunks:
            chunk_size = len(chunk)
            chunk_ranges.append((current_pos, current_pos + chunk_size))
            current_pos += chunk_size
    bytestring = b"".join(chunks)

    return bytestring, chunk_ranges



"""
umbi.io: Utilities for binary and JSON serialization, and tar archive handling.
"""

from .bytes import bytes_to_vector, vector_to_bytes
from .json import *
from .tar import TarReader, TarWriter

__all__ = [
    "bytes_to_vector",
    "JsonLike",
    "json_remove_none_dict_values",
    "json_to_string",
    "TarReader",
    "TarWriter",
]



import os
import tomllib

from .index import UmbIndex
from .logger import set_log_level, setup_logging
from .umb import ExplicitUmb, read_umb, write_umb

__all__ = [
    "set_log_level",
    "setup_logging",
    "UmbIndex",
    "ExplicitUmb",
    "read_umb",
    "write_umb",
]

def get_pyproject_attribute(attribute, default):
    """Read an attribute from pyproject.toml."""
    pyproject_path = os.path.join(os.path.dirname(__file__), "..", "pyproject.toml")
    try:
        with open(pyproject_path, "rb") as f:
            project_data = tomllib.load(f)["project"]
            return project_data.get(attribute)
    except (FileNotFoundError, KeyError):
        return default


__toolname__ = get_pyproject_attribute("name", "unknown")
__version__ = get_pyproject_attribute("version", "unknown")
__format_version__ = get_pyproject_attribute("format_version", 0)
__format_revision__ = get_pyproject_attribute("format_revision", 1)



import logging
from typing import Optional

logger = logging.getLogger(__name__)

from enum import Enum
from types import SimpleNamespace

import umbi
import umbi.io


class UmbFile(Enum):
    """A list of common files expected in a umbfile. Each entry is a tuple of (filename, filetype)."""

    INDEX_JSON = ("index.json", "json")

    INITIAL_STATES = ("initial-states.bin", "vector[bool]")
    STATE_TO_CHOICE = ("state-to-choice.bin", "csr")
    STATE_TO_PLAYER = ("state-to-player.bin", "vector[uint32]")

    MARKOVIAN_STATES = ("markovian-states.bin", "vector[bool]")
    STATE_TO_EXIT_RATE = ("state-to-exit-rate.bin", "csr")
    EXIT_RATES = ("exit-rates.bin", "bytes")

    CHOICE_TO_BRANCH = ("choice-to-branch.bin", "csr")
    CHOICE_TO_ACTION = ("choice-to-action.bin", "vector[uint32]")
    ACTION_TO_ACTION_STRING = ("action-to-action-string.bin", "csr")
    ACTION_STRINGS = ("action-strings.bin", "bytes")

    BRANCH_TO_TARGET = ("branch-to-target.bin", "vector[uint64]")
    BRANCH_TO_PROBABILITY = ("branch-to-probability.bin", "csr")
    BRANCH_PROBABILITIES = ("branch-probabilities.bin", "bytes")


class ExplicitUmb:
    """Class for an explicit representation of a umbfile. The goal of this class is to have all the data is stored in numpy arrays and lists."""

    def __init__(self):
        self.index: umbi.UmbIndex = umbi.UmbIndex()

        self.initial_states = None
        self.state_to_choice = None
        self.state_to_player = None

        self.markovian_states = None
        self.state_exit_rate = None

        self.choice_to_branch = None
        self.choice_to_action = None
        self.action_strings = None

        self.branch_to_target = None
        self.branch_probabilities = None

        self.rewards = None
        self.aps = None
        self.state_valuations = None

    def validate(self):
        # TODO implement
        # logger.debug("UMB validation requested but is not implemented yet")
        return


class UmbReader(umbi.io.TarReader):
    def __init__(self, tarpath: str):
        super().__init__(tarpath)
        # to keep track of which files were read
        self.filename_read = {filename: False for filename in self.filenames}

    def warn_about_unread_files(self):
        """Print warning about unread files from the tarfile, if such exist."""
        unread_files = [f for f, read in self.filename_read.items() if not read]
        for f in unread_files:
            logger.warning(f"umbfile contains unrecognized file: {f}")

    def read_file(self, filename: str, required: bool = False) -> bytes | None:
        """Read raw bytes from a specific file in the tarball. Mark the file as read."""
        logger.debug(f"attempting to read {filename} ...")
        if filename in self.filenames:
            self.filename_read[filename] = True
        return super().read_file(filename, required)

    def read_common(self, file: UmbFile, required: bool = False):
        filename, filetype = file.value
        return self.read_filetype(filename, filetype, required)

    def read_common_csr(self, file: UmbFile, value_type: str, required: bool, file_csr: UmbFile, required_csr: bool = False):
        filename, _ = file.value
        filename_csr, _ = file_csr.value
        return self.read_filetype_csr(filename, value_type, required, filename_csr, required_csr)

    # TODO make annotation into a dedicated class
    def read_annotation(self, label: str, annotation_dict: dict[str, SimpleNamespace] | None):
        """
        Read annotation files. The values from the files will be stored in the corresponding annotation object as a new dictionary mapping applies-to -> values
        :param label: annotation label, usually one of ["rewards","aps"]
        :param annotation_dict: a dictionary annotation name -> annotation
        """
        if annotation_dict is None:
            return
        for name, annotation in annotation_dict.items():
            annotation.values = dict[str, list]()
            for applies in annotation.applies_to:
                path = f"annotations/{label}/{name}/for-{applies}"
                vector = self.read_filetype_csr(
                    f"{path}/values.bin", annotation.type, required=True, filename_csr=f"{path}/to-value.bin"
                )
                assert vector is not None
                annotation.values[applies] = vector

    def read_state_valuations(self, state_valuations: dict[str, object] | None):
        # TODO implement
        if state_valuations is None:
            return
        logger.warning("state valuations import is not implemented yet")
        return

    def read_umb(self):  # type: ignore
        logger.info(f"loading umbfile from {self.tarpath} ...")
        umb = ExplicitUmb()

        json_obj = self.read_common(UmbFile.INDEX_JSON, required=True)
        umb.index = umbi.UmbIndex.from_json(json_obj)

        umb.initial_states = self.read_common(UmbFile.INITIAL_STATES, required=True)
        umb.state_to_choice = self.read_common(UmbFile.STATE_TO_CHOICE)
        umb.state_to_player = self.read_common(UmbFile.STATE_TO_PLAYER)

        umb.markovian_states = self.read_common(UmbFile.MARKOVIAN_STATES)
        if umb.index.transition_system.exit_rate_type is not None:
            umb.state_exit_rate = self.read_common_csr(
                UmbFile.EXIT_RATES, umb.index.transition_system.exit_rate_type, False, UmbFile.STATE_TO_EXIT_RATE, False
            )

        umb.choice_to_branch = self.read_common(UmbFile.CHOICE_TO_BRANCH)
        umb.choice_to_action = self.read_common(UmbFile.CHOICE_TO_ACTION)
        umb.action_strings = self.read_common_csr(UmbFile.ACTION_STRINGS, "string", False, UmbFile.ACTION_TO_ACTION_STRING, True)

        umb.branch_to_target = self.read_common(UmbFile.BRANCH_TO_TARGET)
        if umb.index.transition_system.branch_probability_type is not None:
            umb.branch_probabilities = self.read_common_csr(
                UmbFile.BRANCH_PROBABILITIES, umb.index.transition_system.branch_probability_type, False,
                UmbFile.BRANCH_TO_PROBABILITY, False
            )

        self.read_annotation("rewards", umb.index.annotations.rewards)
        self.read_annotation("aps", umb.index.annotations.aps)
        self.read_state_valuations(umb.index.annotations.state_valuations)

        self.warn_about_unread_files()
        umb.validate()
        logger.info(f"finished loading the umbfile")
        return umb


class UmbWriter(umbi.io.TarWriter):

    def add_common(self, file: UmbFile, data, required: bool = False):
        filename, filetype = file.value
        self.add_filetype(filename, filetype, data, required=required)

    def add_common_csr(self, file: UmbFile, data, value_type: str, file_csr: UmbFile, required: bool = False):
        filename, _ = file.value
        filename_csr, _ = file_csr.value
        self.add_filetype_csr(filename, value_type, data, required, filename_csr)

    def add_annotation(self, label: str, annotation_dict: dict[str, SimpleNamespace] | None):
        """
        Read annotation files.
        :param label: annotation label, usually one of ["rewards","aps"]
        :param annotation_dict: a dictionary annotation name -> annotation
        """
        if annotation_dict is None:
            return
        for name, annotation in annotation_dict.items():
            for applies, values in annotation.values.items():
                prefix = f"annotations/{label}/{name}/for-{applies}"
                self.add_filetype_csr(
                    f"{prefix}/values.bin", annotation.type, values, required=True, filename_csr=f"{prefix}/to-values.bin"
                )

    def add_state_valuations(self, state_valuations: dict[str, list[float]] | None):
        if state_valuations is None:
            return
        logger.warning("state valuations export is not implemented yet")
        return

    def write_umb(self, umb: ExplicitUmb, umbpath: str):
        umb.validate()
        logger.info(f"writing umbfile to {umbpath} ...")
        self.add_common(UmbFile.INDEX_JSON, umb.index.to_json(), required=True)
        self.add_common(UmbFile.INITIAL_STATES, umb.initial_states, required=True)
        self.add_common(UmbFile.STATE_TO_CHOICE, umb.state_to_choice)
        self.add_common(UmbFile.STATE_TO_PLAYER, umb.state_to_player)
        self.add_common(UmbFile.MARKOVIAN_STATES, umb.markovian_states)
        self.add_common_csr(
            UmbFile.EXIT_RATES,
            umb.state_exit_rate,
            umb.index.transition_system.exit_rate_type,
            UmbFile.STATE_TO_EXIT_RATE,
        )
        self.add_common(UmbFile.CHOICE_TO_BRANCH, umb.choice_to_branch)
        self.add_common(UmbFile.CHOICE_TO_ACTION, umb.choice_to_action)
        self.add_common_csr(UmbFile.ACTION_STRINGS, umb.action_strings, "string", UmbFile.ACTION_TO_ACTION_STRING)
        self.add_common(UmbFile.BRANCH_TO_TARGET, umb.branch_to_target)
        self.add_common_csr(
            UmbFile.BRANCH_PROBABILITIES,
            umb.branch_probabilities,
            umb.index.transition_system.branch_probability_type,
            UmbFile.BRANCH_TO_PROBABILITY,
        )
        self.add_annotation("rewards", umb.index.annotations.rewards)
        self.add_annotation("aps", umb.index.annotations.aps)
        self.add_state_valuations(umb.index.annotations.state_valuations)
        self.write(umbpath)
        logger.info(f"finished writing the umbfile")


def read_umb(umbpath: str) -> ExplicitUmb:
    """Read UMB from a umbfile."""
    return UmbReader(umbpath).read_umb()


def write_umb(umb: ExplicitUmb, umbpath: str):
    """Write UMB to a umbfile."""
    UmbWriter().write_umb(umb, umbpath)
